//! This module helps to deal with controls.

use crate::engine::event::{Event, Key, Keyboard};
use crate::engine::vector::Vector;
use crate::log;

/// The different actions that a user can do.
pub enum Action {
    /// The jump button.
    ///
    /// A bool at true means that the button was pressed,
    /// A bool at false means that the button was released.
    Jump(bool),
}

pub enum Controls {
    /// A keyboard controller.
    Keyboard(KeyboardMap),

    /// A gamepad controller.
    Gamepad(GamepadMap),
}

impl Controls {
    /// Returns the default keyboard controls.
    pub fn default_keyboard() -> Controls {
        Controls::Keyboard(KeyboardMap::default())
    }

    /// Returns the default gamepad controls from id.
    ///
    /// Returns None if the gamepad corresponding to the id is not connected.
    pub fn default_gamepad(gamepad: web_sys::Gamepad) -> Controls {
        Controls::Gamepad(GamepadMap::from_gamepad(gamepad))
    }

    // /// Returns the default gamepad controls from id.
    // ///
    // /// Returns None if the gamepad corresponding to the id is not connected.
    // pub fn all_gamepads() -> Vec<Controls> {
    //     GamepadMap::all()
    //         .into_iter()
    //         .map(|x| Controls::Gamepad(x))
    //         .collect()
    // }

    /// Converts an event and depending on the config, returns the corresponding action.
    pub fn convert(&self, event: &Event) -> Option<Action> {
        match self {
            Controls::Keyboard(ref map) => map.convert(event),
            Controls::Gamepad(ref map) => map.convert(event),
        }
    }

    /// Returns the direction of the controls.
    pub fn direction(&self, keyboard: &Keyboard) -> Vector {
        match self {
            Controls::Keyboard(ref map) => map.direction(keyboard),
            Controls::Gamepad(ref map) => map.direction(),
        }
    }

    /// Returns whether the running key is pressed.
    pub fn is_running(&self) -> bool {
        false
        // match self {
        //     Controls::Keyboard(ref map) => map.is_running(),
        //     Controls::Gamepad(ref map) => map.is_running(),
        // }
    }
}

/// A map between keyboard keys and actions.
#[derive(Copy, Clone)]
pub struct KeyboardMap {
    /// The key corresponding to the jump button.
    jump_key: Key,

    /// The key corresponding to the run button.
    run_key: Key,

    /// The key corresponding to the left button.
    left_key: Key,

    /// The key corresponding to the right button.
    right_key: Key,
}

impl KeyboardMap {
    /// Creates the default keyboard config.
    pub fn default() -> KeyboardMap {
        KeyboardMap {
            jump_key: Key::Space,
            run_key: Key::ArrowUp,
            left_key: Key::ArrowLeft,
            right_key: Key::ArrowRight,
        }
    }

    /// Converts an event and depending on the config, returns the corresponding action.
    pub fn convert(&self, event: &Event) -> Option<Action> {
        match event {
            Event::KeyPressed(code) if *code == self.jump_key => Some(Action::Jump(true)),
            Event::KeyReleased(code) if *code == self.jump_key => Some(Action::Jump(false)),
            _ => None,
        }
    }

    /// Returns the direction of the keys.
    pub fn direction(&self, keyboard: &Keyboard) -> Vector {
        let mut ret = Vector::new(0.0, 0.0);
        const RIGHT: Vector = Vector { x: 1.0, y: 0.0 };

        if keyboard.is_key_pressed(self.right_key) {
            ret += RIGHT;
        }
        if keyboard.is_key_pressed(self.left_key) {
            ret -= RIGHT;
        }
        ret
    }
}

/// A map between gamepad buttons and actions.
#[derive(Clone)]
pub struct GamepadMap {
    /// Id of the gamepad.
    inner: web_sys::Gamepad,

    /// Number of the jump button.
    jump_button: u32,

    /// Number of the run button.
    run_button: u32,
    // /// Left / Right axis.
    // left_right_axis: Axis,
}

impl GamepadMap {
    /// Creates a gamepad map from a javascript gamepad.
    pub fn from_gamepad(gamepad: web_sys::Gamepad) -> GamepadMap {
        GamepadMap {
            inner: gamepad,
            jump_button: 0,
            run_button: 3,
        }
    }

    /// Converts an event and depending on the config, returns the corresponding action.
    pub fn convert(&self, event: &Event) -> Option<Action> {
        None
        // match event {
        //     Event::JoystickButtonPressed { joystickid, button }
        //         if *joystickid == self.id && *button == self.jump_button =>
        //     {
        //         Some(Action::Jump(true))
        //     }

        //     Event::JoystickButtonReleased { joystickid, button }
        //         if *joystickid == self.id && *button == self.jump_button =>
        //     {
        //         Some(Action::Jump(false))
        //     }

        //     _ => None,
        // }
    }

    /// Returns the direction of the directionnal buttons of the gamepad.
    pub fn direction(&self) -> Vector {
        Vector::new(0.0, 0.0)
        // Vector2::new(axis_position(self.id, self.left_right_axis) / 100.0, 0.0)
    }

    // /// Returns whether the run button is held down.
    // pub fn is_running(&self) -> bool {
    //     is_button_pressed(self.id, self.run_button)
    // }
}
